package org.naokishibata.math; 

import org.naokishibata.generic.*;

public class ConjGrad {
    public interface Instance extends ScalarField {
	public double value(VectorN v);
	public VectorN gradient(VectorN v);
	public VectorN gradient2(VectorN v);
	public boolean finish(VectorN v);
    }

    static VectorN lineSearch(Instance ins, VectorN iv, VectorN dir, double scale0, double scale1) {
	dir = dir.normalized();
	if (dir.dot(ins.gradient(iv)) > 0) dir = dir.times(-1);

	double a = 0, b = scale0;

	VectorN xa = iv.plus(dir.times(a));
	VectorN xb = iv.plus(dir.times(b));

	double ya = ins.value(xa);
	double yb = ins.value(xb);

	/*
	for(int i=0;i<10;i++) {
	    if (dir.dot(ins.gradient(xb)) <= 0) break;

	    b /= 2;
	    scale1 /= 2;
	    xb = iv.plus(dir.times(b));
	}
	*/

	for(int i=0;;i++) {
	    if (dir.dot(ins.gradient(xb)) >= 0) break;
	    if (yb > ya) break;

	    b *= 2;
	    xb = iv.plus(dir.times(b));
	    yb = ins.value(xb);
	}

	VectorN yad = ins.gradient(xa);
	VectorN ybd = ins.gradient(xb);

	while(xb.minus(xa).length() > scale1) {
	    double c = (a+b) / 2;
	    VectorN xc = iv.plus(dir.times(c));
	    double yc = ins.value(xc);
	    VectorN ycd = ins.gradient(xc);

	    if (dir.dot(ycd) >= 0) {
		b = c;
		xb = xc;
		yb = yc;
		ybd = ycd;
		continue;
	    }

	    if (yc <= ya) {
		a = c;
		xa = xc;
		ya = yc;
		yad = ycd;
		continue;
	    }

	    b = c;
	    xb = xc;
	    yb = yc;
	    ybd = ycd;
	}

	return ya < yb ? xa : xb;
    }

    static VectorN inversedot(VectorN m, VectorN r) {
	double[] a = new double[m.m()];

	boolean isAllPositive = true;

	for(int i=0;i<m.m();i++) {
	    if (m.v(i) <= 0) isAllPositive = false;
	}

	if (isAllPositive) {
	    for(int i=0;i<m.m();i++) {
		a[i] = 1.0 / m.v(i);
		a[i] *= r.v(i);
	    }
	} else return r;

	return new VectorN(a);
    }

    public static VectorN execute(Instance ins, VectorN iv, double scale0, double scale1, int loopCnt) {
	int i = 0, k = 0;
	VectorN x = iv;
	VectorN r = ins.gradient(x).times(-1);
	VectorN m = ins.gradient2(x);
	VectorN s = inversedot(m, r);
	VectorN d = s;
	double deltanew = r.dot(d);
	double delta0 = deltanew;

	while(i < loopCnt) {
	    x = lineSearch(ins, x, d, scale0, scale1);

	    r = ins.gradient(x).times(-1);
	    double deltaold = deltanew;
	    double deltamid = r.dot(s);
	    m = ins.gradient2(x);
	    s = inversedot(m, r);
	    //s = r;
	    deltanew = r.dot(s);
	    double beta = (deltanew - deltamid) / deltaold;
	    if (k == 10 || beta <= 0 || deltaold == 0) {
		d = s;
		k = 0;
	    } else {
		d = s.plus(d.times(beta));
	    }

	    k++;
	    i++;
	}

	return x;
    }

    //

    public static class TestInst implements ConjGrad.Instance {
	public double value(VectorN v) {
	    double x = v.v(0);
	    double y = v.v(1);
	    //return Math.cos(x*x - 3*y) + Math.sin(x*x + y*y);
	    return Math.abs(x) + Math.abs(y);
	    //return x*x + 2*y*y;
	}

	public VectorN gradient(VectorN v) {
	    return MathUtils.gradientAprx(this, v, 0.01);
	}

	public VectorN gradient2(VectorN v) {
	    return MathUtils.gradient2Aprx(this, v, 0.01);
	}

	public boolean finish(VectorN v) {
	    return gradient(v).length() < 0.01;
	}
    }

    public static void main(String[] args) {
	TestInst ti = new TestInst();

	VectorN x0 = ConjGrad.execute(ti, new VectorN(2, 1), 0.001, 0.0001, 20);

	System.out.println(x0);
    }
}
