package org.naokishibata.math;

import java.io.*;
import org.naokishibata.generic.*;

/**
   Matrix is an immutable class representing a matrix
*/
public class Matrix implements Serializable {
    protected final int n,m;
    protected final double a[];

    public Matrix(int m0, int n0) {
	n = n0; m = m0;

	a = new double[m*n];
    }

    public Matrix(Matrix m0) {
	n = m0.n; m = m0.m;

	a = new double[m*n];

	for(int i=0;i<m*n;i++) a[i] = m0.a[i];
    }

    public Matrix(double[][] ary) {
	m = ary.length;
	n = ary[0].length;

	a = new double[m*n];
	for(int i=0;i<m;i++) {
	    for(int j=0;j<n;j++) a[i*n+j] = ary[i][j];
	}
    }

    public Matrix(double[] ary, int m, int n) {
	this.m = m;
	this.n = n;

	a = new double[m*n];
	for(int i=0;i<m*n;i++) {
	    a[i] = ary[i];
	}
    }

    public String toString() {
	String s = new String();

	for(int i=0;i<m;i++) {
	    for(int j=0;j<n;j++) {
		s = s + a[i*n+j];
		if (j != n-1) s = s + ", ";
	    }
	    s = s + "\n";
	}	

	return s;
    }

    public boolean equals(Matrix mat) {
	if (n != mat.n || m != mat.m) return false;

	for(int i=0;i<m*n;i++) {
	    if (a[i] != mat.a[i]) return false;
	}
	    
	return true;
    }

    public boolean equals(Object o) {
	return equals((Matrix)o);
    }

    public int hashCode() {
	int ret = 0;

	for(int i=0;i<m;i++) {
	    for(int j=0;j<n;j++) {
		ret |= (int)(Double.doubleToRawLongBits(a[i*n+j] * m + n));
	    }
	}

	return ret;
    }

    public static Matrix identity(int n0) {
	Matrix ret = new Matrix(n0,n0);

	for(int i=0;i<n0;i++) {
	    for(int j=0;j<n0;j++) ret.a[i*n0+j] = i == j ? 1.0 : 0.0;
	}

	return ret;
    }

    /*
    public void loadIdentity() {
	for(int i=0;i<n;i++) {
	    for(int j=0;j<m;j++) a[i*n+j] = i == j ? 1.0 : 0.0;
	}
    }
    */

    public double v(int y, int x) {
	return a[y*n+x];
    }

    public Matrix put(int y, int x, double d) {
	Matrix m = new Matrix(this);
	m.a[y*n+x] = d;
	return m;
    }

    public Matrix put(int y, int x, Matrix s) {
	assert y + s.m <= m && x + s.n <= n;

	Matrix r = new Matrix(this);

	for(int i=0;i<s.m;i++) {
	    for(int j=0;j<s.n;j++) {
		r.a[(y+i)*n+x+j] = s.a[i*n+j];
	    }
	}

	return r;
    }

    public int n() {
	return n;
    }

    /**
       Returns the number of of this matrix
    */
    public int m() {
	return m;
    }

    public Matrix times(Matrix m0) {
	assert n == m0.m : "n = " + n + ", m0.m = " + m0.m;

	Matrix r = new Matrix(m, m0.n);

	for(int i=0;i<m;i++) {
	    for(int j=0;j<m0.n;j++) {
		r.a[i*r.n+j] = 0;
		for(int k=0;k<n;k++) {
		    r.a[i*r.n+j] += a[i*n+k] * m0.a[k*m0.n+j];
		}
	    }
	}

	return r;
    }

    public Matrix times(double v) {
	Matrix r = new Matrix(this);

	for(int i=0;i<m*n;i++) r.a[i] *= v;

	return r;
    }

    public Matrix plus(Matrix m0) {
	assert n == m0.n && m == m0.m;

	Matrix r = new Matrix(this);

	for(int i=0;i<m*n;i++) r.a[i] = a[i] + m0.a[i];

	return r;
    }

    public Matrix minus(Matrix m0) {
	assert n == m0.n && m == m0.m;

	Matrix r = new Matrix(this);

	for(int i=0;i<m*n;i++) r.a[i] = a[i] - m0.a[i];

	return r;
    }

    /**
       Returns the transposition of this matrix
    */
    public Matrix transposition() {
	Matrix r = new Matrix(n, m);

	for(int i=0;i<m;i++) {
	    for(int j=0;j<n;j++) r.a[i*n+j] = a[j*n+i];
	}

	return r;
    }

    private double lusub(int[] ip0) {
	int i, j=0, ii, ik;
	double t, u, det;
	double w[] = new double[n];
	int ip[] = new int[n];

	det = 0;
	for (int k = 0; k < n; k++) {
	    ip[k] = k;
	    u = 0;
	    for (j = 0; j < n; j++) {
		t = Math.abs(a[k*n+j]);
		if (t > u) u = t;
	    }
	    if (u == 0) return 0;
	    w[k] = 1 / u;
	}
	det = 1;
	for (int k = 0; k < n; k++) {
	    u = -1;
	    for (i = k; i < n; i++) {
		ii = ip[i];
		t = Math.abs(a[ii*n+k]) * w[ii];
		if (t > u) {  u = t;  j = i;  }
	    }
	    ik = ip[j];
	    if (j != k) {
		ip[j] = ip[k];  ip[k] = ik;
		det = -det;
	    }
	    u = a[ik*n+k];  det *= u;
	    if (u == 0) return 0;
	    for (i = k + 1; i < n; i++) {
		ii = ip[i];
		t = (a[ii*n+k] /= u);
		for (j = k + 1; j < n; j++)
		    a[ii*n+j] -= t * a[ik*n+j];
	    }
	}

	for(i=0;i<n;i++) ip0[i] = ip[i];

	return det;
    }

    /**
       Returns the inversion of this matrix
    */
    public Matrix inversion() {
	assert n == m;

	Matrix ret = identity(n);
	Matrix lu  = new Matrix(this);
	int ii;
	double t;
	int ip[] = new int[n];

	double det = lu.lusub(ip);

	if (det != 0) {
	    for (int k = 0; k < n; k++) {
		for (int i = 0; i < n; i++) {
		    ii = ip[i];
		    t = ii == k ? 1 : 0;
		    for (int j = 0; j < i; j++)
			t -= lu.a[ii*n+j] * ret.a[j*n+k];
		    ret.a[i*n+k] = t;
		}
		for (int i = n - 1; i >= 0; i--) {
		    t = ret.a[i*n+k];
		    ii = ip[i];
		    for (int j = i + 1; j < n; j++)
			t -= lu.a[ii*n+j] * ret.a[j*n+k];
		    ret.a[i*n+k] = t / lu.a[ii*n+i];
		}
	    }
	}

	return ret;
    }

    /**
       Returns the determinant of this matrix
    */
    public double determinant() {
	return (new Matrix(this)).lusub(new int[n]);
    }
}
