package org.naokishibata.armarker;

import java.util.*;
import java.nio.charset.*;

import static org.jocl.CL.*;
import org.jocl.*;

import org.naokishibata.generic.*;

/**
 * This is a helper class for OpenCL.
 */
public class CLHelper {
    final cl_platform_id platform;
    final cl_device_id device;

    final cl_context context;
    final cl_command_queue commandQueue;

    public static class PlatformDevicePair {
	public final cl_platform_id platform;
	public final cl_device_id device;
	public final cl_context_properties contextProperties;

	PlatformDevicePair(cl_platform_id platform, cl_device_id device, cl_context_properties contextProperties) {
	    this.platform = platform;
	    this.device = device;
	    this.contextProperties = contextProperties;
	}

	public String toString() {
	    String s = "";
	    s = s + "Platform : " + getString(platform, CL_PLATFORM_NAME) + "\n";
	    s = s + "Device : " + getString(device, CL_DEVICE_NAME) + "\n";
	    s = s + "Version : " + getString(device, CL_DRIVER_VERSION) + "\n";
	    return s;
	}

	public boolean equals(Object o) {
	    if (o == this) return true;
	    if (!(o instanceof PlatformDevicePair)) return false;
	    PlatformDevicePair p = (PlatformDevicePair) o;
	    if (p.platform.equals(platform) && p.device.equals(device)) return true;
	    return false;
	}

	public int hashCode() {
	    return platform.hashCode() ^ device.hashCode();
	}
    }

    public static PlatformDevicePair[] listPlatformDevice() {
        int numPlatformsArray[] = new int[1];
        clGetPlatformIDs(0, null, numPlatformsArray);
        int numPlatforms = numPlatformsArray[0];

	LinkedHashSet<PlatformDevicePair> pairset = new LinkedHashSet<PlatformDevicePair>();

	cl_platform_id[] platforms = new cl_platform_id[numPlatforms];
        clGetPlatformIDs(platforms.length, platforms, null);

	for(int platformIndex = 0;platformIndex < numPlatforms;platformIndex++) {
	    cl_platform_id platform = platforms[platformIndex];

	    cl_context_properties contextProperties = new cl_context_properties();
	    contextProperties.addProperty(CL_CONTEXT_PLATFORM, platform);

	    int numDevicesArray[] = new int[1];
	    clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 0, null, numDevicesArray);
	    int numDevices = numDevicesArray[0];

	    if (numDevices == 0) continue;

	    cl_device_id devices[] = new cl_device_id[numDevices];
	    clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, numDevices, devices, null);

	    for(int i=0;i<numDevices;i++) {
		pairset.add(new PlatformDevicePair(platform, devices[i], contextProperties));
	    }
	}

	for(int platformIndex = 0;platformIndex < numPlatforms;platformIndex++) {
	    cl_platform_id platform = platforms[platformIndex];

	    cl_context_properties contextProperties = new cl_context_properties();
	    contextProperties.addProperty(CL_CONTEXT_PLATFORM, platform);

	    int numDevicesArray[] = new int[1];
	    clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, 0, null, numDevicesArray);
	    int numDevices = numDevicesArray[0];

	    if (numDevices == 0) continue;

	    cl_device_id devices[] = new cl_device_id[numDevices];
	    clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, numDevices, devices, null);

	    for(int i=0;i<numDevices;i++) {
		pairset.add(new PlatformDevicePair(platform, devices[i], contextProperties));
	    }
	}

	return pairset.toArray(new PlatformDevicePair[0]);
    }

    public CLHelper(int deviceIndex, long prop) {
	PlatformDevicePair[] pairs = listPlatformDevice();

	if (deviceIndex >= pairs.length) {
	    Debug.abort("Device ID " + deviceIndex + " does not exist");
	}

	platform = pairs[deviceIndex].platform;
        device   = pairs[deviceIndex].device;

        context = clCreateContext(pairs[deviceIndex].contextProperties, 1, new cl_device_id[]{device}, null, null, null);
        
        // Create a command-queue for the selected device
        //commandQueue = clCreateCommandQueue(context, device, CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE, null);
        commandQueue = clCreateCommandQueue(context, device, prop, null);
    }

    public CLHelper(int deviceIndex) {
	this(deviceIndex, 0);
    }

    public CLHelper(CLHelper org, long prop) {
	platform = org.platform;
	device = org.device;
	context = org.context;

        commandQueue = clCreateCommandQueue(context, device, prop, null);
    }

    public cl_kernel compile(String kernelName, String progString, String options) {
	int[] t = new int[1];

	cl_program program = clCreateProgramWithSource(context, 1, new String[]{ progString }, null, t);

	if (t[0] != CL_SUCCESS) {
	    System.out.println("clCreateProgramWithSource terminated with error code " + t[0]);
	    System.exit(-1);
	}


        int ret = clBuildProgram(program, 0, null, options, null, null);

	byte[] strbuf = new byte[1030];

	if (ret != CL_SUCCESS) {
	    if (clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG, 1024, Pointer.to(strbuf), null) == CL_SUCCESS) {
		System.out.println(new String(strbuf, Charset.forName("US-ASCII")));
	    }

	    System.out.println("clBuildProgram not succeeded :" + ret);
	    System.exit(-1);
	}

	int[] errcode_ret = new int[1];

        cl_kernel kernel = clCreateKernel(program, kernelName, errcode_ret);
	if (errcode_ret[0] != CL_SUCCESS) {
	    System.out.println("clBuildProgram not succeeded : " + errcode_ret[0]);
	    System.exit(-1);
	}

        clReleaseProgram(program);

	return kernel;
    }

    public int run1D(cl_kernel kernel, long gSize, long lSize) {
        long global_work_size[] = new long[]{gSize};
        long local_work_size[] = null;

	if (lSize > 0) local_work_size = new long[]{lSize};
        
        return clEnqueueNDRangeKernel(commandQueue, kernel, 1, null, global_work_size, local_work_size, 0, null, null);
    }

    public int run2D(cl_kernel kernel, long gSizeX, long gSizeY, long lSizeX, long lSizeY) {
        long global_work_size[] = new long[]{gSizeX, gSizeY};
        long[] local_work_size = null;

	if (lSizeX > 0) local_work_size = new long[]{lSizeX, lSizeY};
        
        return clEnqueueNDRangeKernel(commandQueue, kernel, 2, null, global_work_size, local_work_size, 0, null, null);
    }

    public int run2D(cl_kernel kernel, long gSizeX, long gSizeY, long lSizeX, long lSizeY, cl_event event, cl_event ... eventsInWait) {
        long global_work_size[] = new long[]{gSizeX, gSizeY};
        long[] local_work_size = null;

	if (lSizeX > 0) local_work_size = new long[]{lSizeX, lSizeY};

	if (eventsInWait.length != 0 && eventsInWait[0] == null) {
	    eventsInWait = new cl_event[0];
	}

	return clEnqueueNDRangeKernel(commandQueue, kernel, 2, null, global_work_size, local_work_size, eventsInWait.length, eventsInWait, event);
    }

    public int run2D(cl_kernel kernel, long offsetX, long offsetY, long gSizeX, long gSizeY, long lSizeX, long lSizeY) {
        long[] offset = new long[]{offsetX, offsetY};
        long[] global_work_size = new long[]{gSizeX, gSizeY};
        long[] local_work_size = null;

	if (lSizeX > 0) local_work_size = new long[]{lSizeX, lSizeY};
        
        return clEnqueueNDRangeKernel(commandQueue, kernel, 2, offset, global_work_size, local_work_size, 0, null, null);
    }

    public int run2D(cl_kernel kernel, long offsetX, long offsetY, long gSizeX, long gSizeY, long lSizeX, long lSizeY, cl_event event, cl_event ... eventsInWait) {
        long[] offset = new long[]{offsetX, offsetY};
        long global_work_size[] = new long[]{gSizeX, gSizeY};
        long[] local_work_size = null;

	if (lSizeX > 0) local_work_size = new long[]{lSizeX, lSizeY};
        
	if (eventsInWait.length != 0 && eventsInWait[0] == null) {
	    eventsInWait = new cl_event[0];
	}

	return clEnqueueNDRangeKernel(commandQueue, kernel, 2, offset, global_work_size, local_work_size, eventsInWait.length, eventsInWait, event);
    }

    public long workGroupSize(cl_kernel kernel) {
	long[] param_value = new long[1];
	clGetKernelWorkGroupInfo(kernel, device, CL_KERNEL_WORK_GROUP_SIZE, Sizeof.size_t, Pointer.to(param_value), null);
	return param_value[0];
    }

    public void flush() {
	clFlush(commandQueue);
    }

    public void finish() {
	clFinish(commandQueue);
    }

    public void dispose() {
        clReleaseCommandQueue(commandQueue);
        clReleaseContext(context);
    }

    private static String getString(cl_device_id device, int paramName) {
        long size[] = new long[1];
        clGetDeviceInfo(device, paramName, 0, null, size);

        byte buffer[] = new byte[(int)size[0]];
        clGetDeviceInfo(device, paramName, buffer.length, Pointer.to(buffer), null);

        return new String(buffer, 0, buffer.length-1);
    }

    private static String getString(cl_platform_id platform, int paramName)
    {
        long size[] = new long[1];
        clGetPlatformInfo(platform, paramName, 0, null, size);

        byte buffer[] = new byte[(int)size[0]];
        clGetPlatformInfo(platform, paramName, buffer.length, Pointer.to(buffer), null);

        return new String(buffer, 0, buffer.length-1);
    }

    public void showInfo() {
	System.out.println("Platform : " + getString(platform, CL_PLATFORM_NAME));
	System.out.println("Device : " + getString(device, CL_DEVICE_NAME));
	System.out.println("Version : " + getString(device, CL_DRIVER_VERSION));
    }

    public cl_context context() { return context; }
    public cl_command_queue commandQueue() { return commandQueue; }

    public static long startTime(cl_event event) {
	long t[] = new long[1];
	CL.clGetEventProfilingInfo(event, CL.CL_PROFILING_COMMAND_START, Sizeof.cl_ulong, Pointer.to(t), null);
	return t[0];
    }

    public static long endTime(cl_event event) {
	long t[] = new long[1];
	CL.clGetEventProfilingInfo(event, CL.CL_PROFILING_COMMAND_END, Sizeof.cl_ulong, Pointer.to(t), null);
	return t[0];
    }

    public static void release(cl_event ... a) {
	for(int i=0;i<a.length;i++) {
	    CL.clReleaseEvent(a[i]);
	}
    }
}
