package org.naokishibata.generic;

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;

/** 
 * This class implements a tiny HTTP server whose responses can be
 * generated by a method given by a user.
 */
public class TinyHttpd implements Runnable {
    static int nWorkers = 2;
    static final byte[] newline = {(byte)0x0d, (byte)0x0a, (byte)0x0d, (byte)0x0a};
    static SimpleDateFormat dateFormat = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz", Locale.US);

    Thread mainThread;
    ServerSocket ss;
    Object mutex = new Object();
    Worker[] worker = new Worker[nWorkers];
    Responder responder;

    /** 
     * This class represents a HTTP response given by this server class
     */
    public static class Response {
	final String statusLine, contentEncoding;
	final byte[] entityBody;

	private Response(String statusLine, String contentEncoding, byte[] entityBody) {
	    this.statusLine = statusLine;

	    if (contentEncoding != null) {
		byte[] ea;
		String ce = null;
		try {
		    ea = contentEncoding.getBytes("US-ASCII");
		    for(int i=0;i<ea.length;i++) {
			if (ea[i] == (byte)0x0d) ea[i] = (byte)' ';
			if (ea[i] == (byte)0x0a) ea[i] = (byte)' ';
		    }
		    ce = new String(ea, "US-ASCII");
		} catch(UnsupportedEncodingException ex) {
		}
		this.contentEncoding = ce;
	    } else {
		this.contentEncoding = null;
	    }

	    if (entityBody != null) {
		this.entityBody = new byte[entityBody.length];
		for(int i=0;i<entityBody.length;i++) {
		    this.entityBody[i] = entityBody[i];
		}
	    } else {
		this.entityBody = null;
	    }
	}

	/** 
	 * Generates a 404 response
	 */
	public static Response response404() { return new Response("404 not found", null, null); }

	/** 
	 * Generates a 405 response
	 */
	public static Response response405() { return new Response("405 not implemented", null, null); }

	/** 
	 * Generates a 200 response
	 * @param encoding Content encoding
	 * @param body Entity body
	 */
	public static Response response200(String encoding, byte[] body) { return new Response("200 OK", encoding, body); }

	/** 
	 * Generates a 204 response
	 */
	public static Response response204() { return new Response("204 no content", null, null); }

	/** 
	 * Generates a 500 response
	 */
	public static Response response500() { return new Response("500 server error", null, null); }
    }

    /** 
     * An interface for a callback method with which the server
     * response can be generated by a method given by the user.
     * @param reqLine The whole request line, e.g. "GET / HTTP/1.1",
     * given by the server.
     * @param method A string, e.g. "GET", indicating the method given
     * by the server.
     * @param method A string, e.g. "/index.html", indicating the
     * requested URI given by the server.
     */
    public interface Responder {
	public Response requestReceived(String reqLine, String method, String uri);
    }

    class Worker implements Runnable {
	Socket s = null;

	void setSocket(Socket s) {
	    synchronized(mutex) {
		this.s = s;
		mutex.notifyAll();
	    }
	}

	public void run() {
	    for(;;) {
		synchronized(mutex) {
		    while(s == null) {
			try { mutex.wait(); } catch (InterruptedException ex) {}
		    }
		}

		PrintStream ps = null;
		LineNumberReader lnr = null;

		try {
		    s.setSoTimeout(3000);
		    s.setTcpNoDelay(true);

		    lnr = new LineNumberReader(new InputStreamReader(s.getInputStream(), "US-ASCII"));
		    ps = new PrintStream(s.getOutputStream());

		    String requestLine = lnr.readLine();

		    String[] a = MiscUtils.regexCapture(requestLine, "([A-Z]+) ([^ ]+) (HTTP/[0-9]+\\.[0-9]+).*");

		    Response res;
		    if (a != null) {
			res = responder.requestReceived(requestLine, a[0], a[1]);
			if (res == null) res = Response.response500();
		    } else {
			res = Response.response405();
		    }

		    if (requestLine.startsWith("GET ") || requestLine.startsWith("HEAD ")) {
			ps.print("HTTP/1.0 " + res.statusLine);
			ps.write(newline, 0, 2);
			ps.print("Server: TinyHttpd.java");
			ps.write(newline, 0, 2);
			ps.print("Date: " + (dateFormat.format(new Date())));
			ps.write(newline, 0, 2);
			ps.print("Access-Control-Allow-Origin: *");
			ps.write(newline, 0, 2);
			if (res.entityBody != null) {
			    ps.print("Content-length: " + res.entityBody.length);
			    ps.write(newline, 0, 2);
			}
			ps.print("Content-type: " + res.contentEncoding);
			ps.write(newline, 0, 4);
			if (res.entityBody != null) {
			    ps.write(res.entityBody, 0, res.entityBody.length);
			}
		    } else {
			ps.print("HTTP/1.0 405 not implemented");
			ps.write(newline, 0, 4);
		    }
		} catch(IOException ex) {
		}

		if (lnr != null) try { lnr.close(); } catch(IOException ex) {}
		ps.close();

		synchronized(mutex) {
		    s = null;
		    mutex.notifyAll();
		}
	    }
	}
    }

    /**
     * Do not call
     */
    public void run() {
        for(;;) {
            Socket s = null;

	    while(s == null) {
		try { s = ss.accept(); } catch(IOException ex) {}
	    }

	    synchronized(mutex) {
		for(int i=0;i<nWorkers;i++) {
		    if (worker[i].s == null) {
			worker[i].setSocket(s);
			break;
		    }
		}

		try { mutex.wait(); } catch (InterruptedException ex) {}
	    }
        }
    }

    /**
     * @param port The port number on which the created server will
     * listen.
     * @param res The responder method which will generate responses from the server.
     */
    public TinyHttpd(int port, Responder res) throws IOException {
        ss = new ServerSocket(port);
	responder = res;

        for (int i = 0;i < nWorkers;i++) {
	    (new Thread(worker[i] = new Worker())).start();
        }

	(mainThread = new Thread(this)).start();
    }

    /**
     * An example implementation of an HTTP server.
     */
    public static void main(String[] args) throws Exception {
	TinyHttpd d = new TinyHttpd(12345, new TinyHttpd.Responder() {
		public Response requestReceived(String reqLine, String method, String uri) {
		    System.out.println("reqLine = " + reqLine);
		    System.out.println("method = " + method);
		    System.out.println("uri = " + uri);
		    return Response.response200("text/plain", uri.getBytes());
		}
	    });
    }
}
