package org.naokishibata.ocvimgio;

import java.util.*;
import java.awt.image.*;
import java.nio.*;

import org.naokishibata.generic.*;

/**
   An instance of this class encapsulates an IplImage instance.
 */
public class IplImage {
    public static final int IPL_DEPTH_SIGN = 0x80000000;
    public static final int IPL_DEPTH_1U = 1;
    public static final int IPL_DEPTH_8U = 8;
    public static final int IPL_DEPTH_16U = 16;
    public static final int IPL_DEPTH_32F = 32;

    public static final int IPL_DEPTH_8S  = (IPL_DEPTH_SIGN| 8);
    public static final int IPL_DEPTH_16S = (IPL_DEPTH_SIGN|16);
    public static final int IPL_DEPTH_32S = (IPL_DEPTH_SIGN|32);

    byte[] pointer;
    int mode;

    private int[] body = null;
    private byte[] imageData = null;

    static final int MODE_NO_MODIFICATION = 1;

    IplImage(byte[] p, int mode) {
	pointer = new byte[p.length];

	for(int i=0;i<p.length;i++) {
	    pointer[i] = p[i];
	}

	this.mode = mode;
    }

    /**
       Make an IplImage instance by calling cvCreateImage.
    */
    public IplImage(int width, int height) {
	pointer = OpenCVJNI.cvCreateImage(width, height, 8, 3);
	refreshBody();
    }

    /**
       Make an IplImage instance by calling cvCreateImage.
    */
    public IplImage(int width, int height, int depth, int nch) {
	pointer = OpenCVJNI.cvCreateImage(width, height, depth, nch);
	refreshBody();
    }

    /**
       Make a clone of IplImage instance by calling cvCloneImage.
    */
    public IplImage(IplImage img) {
	pointer = OpenCVJNI.cvCloneImage(img.pointer);
	refreshBody();
    }

    private void refreshBody() {
	if (body == null) {
	    synchronized(this) {
		if (body == null && pointer != null) {
		    body = OpenCVJNI.decodeIplImage(pointer);
		}
	    }
	}
    }

    private void refreshData() {
	if (imageData == null && pointer != null) {
	    synchronized(this) {
		if (imageData == null && pointer != null) {
		    imageData = OpenCVJNI.imageDataIplImage(pointer);
		}
	    }
	}
    }

    public int nSize() {
	refreshBody();
	return body[0];
    }

    public int ID() {
	refreshBody();
	return body[1];
    }

    public int nChannels() {
	refreshBody();
	return body[2];
    }

    public int depth() {
	refreshBody();
	return body[4];
    }

    public int dataOrder() {
	refreshBody();
	return body[5];
    }

    public int origin() {
	refreshBody();
	return body[6];
    }

    public int align() {
	refreshBody();
	return body[7];
    }

    public int width() {
	refreshBody();
	return body[8];
    }

    public int height() {
	refreshBody();
	return body[9];
    }

    public int imageSize() {
	refreshBody();
	return body[10];
    }

    public int widthStep() {
	refreshBody();
	return body[11];
    }

    public int colorMode() {
	refreshBody();
	return body[12];
    }

    public int channelSeq() {
	refreshBody();
	return body[13];
    }

    /**
       Returns an ByteBuffer instance that points the imageData of this IplImage object.
    */
    public ByteBuffer getDirect() {
	synchronized(this) {
	    if (pointer != null) return OpenCVJNI.imageDataIplImageDirect(pointer);
	}
	return null;
    }

    /**
       Copies the current image to a BufferedImage instance and return it.
    */
    public BufferedImage getImage() {
	int w = width(), h = height(), ws = widthStep();

	//BufferedImage img = BufferedImagePool.getBufferedImage(w, h, BufferedImage.TYPE_3BYTE_BGR);
	BufferedImage img = new BufferedImage(w, h, BufferedImage.TYPE_3BYTE_BGR);

	if (nChannels() != 3) Debug.abort("nChannels != 3");
	if (dataOrder() != 0) Debug.abort("dataOrder != 0");
	if (depth() != 8) Debug.abort("depth != 8");

	if (imageSize() != widthStep()*height()) Debug.abort("imageSize() != widthStep()*height()");

	refreshData();

	int dy = origin() == 0 ? 0 : h-1;
	int dyi = origin() == 0 ? 1 : -1;

	if (w % 8 != 0) {
	    for(int y=0;y<h;y++) {
		for(int x=0;x<w;x++) {
		    int r = imageData[ws*y + x*3 + 0] & 0xff;
		    int g = imageData[ws*y + x*3 + 1] & 0xff;
		    int b = imageData[ws*y + x*3 + 2] & 0xff;
		    int rgb = (b << 16) | (g << 8) | r;

		    img.setRGB(x, dy, rgb);
		}
		dy += dyi;
	    }
	} else {
	    int[] a = new int[8];

	    for(int y=0;y<h;y++) {
		for(int x=0;x<w;x+=8) {
		    a[0] = ((imageData[ws*y + x*3 +  2] & 0xff) << 16) | ((imageData[ws*y + x*3 +  1] & 0xff) << 8) | (imageData[ws*y + x*3 +  0] & 0xff);
		    a[1] = ((imageData[ws*y + x*3 +  5] & 0xff) << 16) | ((imageData[ws*y + x*3 +  4] & 0xff) << 8) | (imageData[ws*y + x*3 +  3] & 0xff);
		    a[2] = ((imageData[ws*y + x*3 +  8] & 0xff) << 16) | ((imageData[ws*y + x*3 +  7] & 0xff) << 8) | (imageData[ws*y + x*3 +  6] & 0xff);
		    a[3] = ((imageData[ws*y + x*3 + 11] & 0xff) << 16) | ((imageData[ws*y + x*3 + 10] & 0xff) << 8) | (imageData[ws*y + x*3 +  9] & 0xff);
		    a[4] = ((imageData[ws*y + x*3 + 14] & 0xff) << 16) | ((imageData[ws*y + x*3 + 13] & 0xff) << 8) | (imageData[ws*y + x*3 + 12] & 0xff);
		    a[5] = ((imageData[ws*y + x*3 + 17] & 0xff) << 16) | ((imageData[ws*y + x*3 + 16] & 0xff) << 8) | (imageData[ws*y + x*3 + 15] & 0xff);
		    a[6] = ((imageData[ws*y + x*3 + 20] & 0xff) << 16) | ((imageData[ws*y + x*3 + 19] & 0xff) << 8) | (imageData[ws*y + x*3 + 18] & 0xff);
		    a[7] = ((imageData[ws*y + x*3 + 23] & 0xff) << 16) | ((imageData[ws*y + x*3 + 22] & 0xff) << 8) | (imageData[ws*y + x*3 + 21] & 0xff);

		    img.setRGB(x, dy, 8, 1, a, 0, 0);
		}
		dy += dyi;
	    }
	}

	return img;
    }

    /**
       Copies a BufferedImage instance to this IplImage instance
    */
    public void setImage(BufferedImage bimg) {
	if ((mode & MODE_NO_MODIFICATION) != 0) return;

	int w = Math.min(bimg.getWidth() , width());
	int h = Math.min(bimg.getHeight(), height());

	ByteBuffer bb = getDirect();

	int[] a = new int[w];

	for(int y=0;y<h;y++) {
	    bimg.getRGB(0, y, w, 1, a, 0, 0);
	    for(int x=0;x<w;x++) {
		bb.put((byte)(a[x] >>  0));
		bb.put((byte)(a[x] >>  8));
		bb.put((byte)(a[x] >> 16));
	    }
	}
    }

    /**
       Release the resource allocated for this IplImage by calling
       cvReleaseImage.
    */
    public void dispose() {
	if ((mode & MODE_NO_MODIFICATION) != 0) return;

	synchronized(this) {
	    if (pointer != null) {
		OpenCVJNI.cvReleaseImage(pointer);
		pointer = null;
	    }
	}
    }
}
